/*************************************************************************
 *
 * FILE  main.cpp
 * Copyright (c) 2005, 2006 Daniel Kahlin <daniel@kahlin.net>
 * Written by Daniel Kahlin <daniel@kahlin.net>
 * $Id: main.cpp,v 1.48 2007-01-02 21:53:43 tlr Exp $
 *
 * DESCRIPTION 
 *
 ******/

#include <iostream>
#include <string>
#include <stdio.h>
#include "util/types.hpp"
#include "util/getopt.hpp"
#include "global.hpp"
#include "boot.hpp"
#include "transfer.hpp"
#include "membuffer.hpp"
// arch dependant
#include "timer.hpp"
#include "port.hpp"

#define PROGRAM "dtvtrans"

#ifndef IS_LINUX
#define PORT_DEFAULT "0x378"
#else
#define PORT_DEFAULT "/dev/parport0"
#endif

typedef int (*cmd_func_t)(Getopt&);
int cmd_boot(Getopt& opts);
int cmd_read_mem(Getopt& opts);
int cmd_write_mem(Getopt& opts);
int cmd_execute_mem(Getopt& opts);
int cmd_clear_port(Getopt& opts);

int main(int argc, char *argv[])
{
    std::map<std::string,cmd_func_t> cmds;
    cmds.insert(std::make_pair<std::string,cmd_func_t>("boot",cmd_boot));
    cmds.insert(std::make_pair<std::string,cmd_func_t>("bt",cmd_boot));
    cmds.insert(std::make_pair<std::string,cmd_func_t>("b",cmd_boot));
    cmds.insert(std::make_pair<std::string,cmd_func_t>("read",cmd_read_mem));
    cmds.insert(std::make_pair<std::string,cmd_func_t>("rd",cmd_read_mem));
    cmds.insert(std::make_pair<std::string,cmd_func_t>("r",cmd_read_mem));
    cmds.insert(std::make_pair<std::string,cmd_func_t>("write",cmd_write_mem));
    cmds.insert(std::make_pair<std::string,cmd_func_t>("wr",cmd_write_mem));
    cmds.insert(std::make_pair<std::string,cmd_func_t>("w",cmd_write_mem));
    cmds.insert(std::make_pair<std::string,cmd_func_t>("go",cmd_execute_mem));
    cmds.insert(std::make_pair<std::string,cmd_func_t>("g",cmd_execute_mem));
    cmds.insert(std::make_pair<std::string,cmd_func_t>("clear",cmd_clear_port));
    cmds.insert(std::make_pair<std::string,cmd_func_t>("clr",cmd_clear_port));
    cmds.insert(std::make_pair<std::string,cmd_func_t>("c",cmd_clear_port));

    /* defaults */
    verbose_g = false;
    debug_g   = false;
    portbase_g = PORT_DEFAULT;
    blocksize_g = 0x0400;
    rawmode_g = false;

    /*
     * First scan for any options.
     *
     */
    Getopt::Opt opt;
    Getopt opts(argc, argv, "rb:p:vdVh");
    while (opt = opts.next_opt(), !opt.is_end()) {

	switch(opt.get_status()) {
	case Getopt::Opt::OK:
	    break;
	case Getopt::Opt::MISSING:
	    fprintf(stderr, PROGRAM ": missing parameter for option -%c\n",opt.get_opt());
	    return 1;
	case Getopt::Opt::INVALID:

	    fprintf(stderr, PROGRAM ": invalid option -%c\n",opt.get_opt());
	    return 1;
	default:
	    fprintf(stderr, PROGRAM ": error while parsing option -%c\n",opt.get_opt());
	    return 1;
	}

	switch (opt.get_opt()) {

        /* set raw format */
        case 'r':
	    rawmode_g = true;
            break;

        case 'b':
	    {
		const char* str=opt.get_arg();
		char* endptr;
		blocksize_g=std::strtoul(str,&endptr,0);
		if (endptr==str) {
		    fprintf(stdout, "invalid number (%s)\n",str);// failed
		    return 1;
		}
	    }
	    if (blocksize_g < 1 || blocksize_g > 0x4000) {
		fprintf(stdout, "blocksize out of range (=%d, allowable 1-16384)\n", blocksize_g );
		return -1;
	    }
            break;

        case 'p':
	    portbase_g=opt.get_arg();
            break;

        /* set verbose */
        case 'v':
	    verbose_g = true;
            break;

        /* set debug */
        case 'd':
	    debug_g = true;
            break;

        /* print version */
        case 'V':
            fprintf (stdout, PROGRAM " " VERSION "\n");
            return 0;

        /* print help */
        case 'h':
            fprintf (stdout,
PROGRAM " " VERSION "\n"
"Copyright (c) 2005, 2006, 2007 Daniel Kahlin\n"
"Written by Daniel Kahlin <daniel@kahlin.net>\n"
#ifdef IS_LINUX
"Linux backend written by Hannu Nuotio <hannu.nuotio@pp.inet.fi>\n"
"Copyright (c) 2006, 2007 Hannu Nuotio\n"
#endif
"\n"
"usage: " PROGRAM " [OPTION]... <cmd> [OPTION]... [<file>]\n"
"\n"
"Valid options:\n"
"    -r              use raw format for files (no load address)\n"
"    -b <size>       block size (default: 1024)\n"
"    -p <addr>       port address (default: " PORT_DEFAULT ")\n"
"    -v              be verbose\n"
"    -d              output debugging information\n"
"    -h              displays this help text\n"
"    -V              output program version\n"
"\n"
"Valid commands:\n"
"    boot/bt/b       send file to boot.prg/mlboot.prg\n"
"    read/rd/r       read mem\n"
"    write/wr/w      write mem\n"
"    go/g            execute mem (go)\n"
"    clear/clr/c     clear port (tristate)\n"
"\n");
            return 0;

        /* default behavior */
        default:
            break;
        }
    }

    if (opts.args_left()<1) {
	printf("too few arguments\n");
	return 1;
    }
    const char* cmd=opts.next_arg();

    int ret=1;
    std::map<std::string,cmd_func_t>::const_iterator it=cmds.find(cmd);
    if (it!=cmds.end()) {
	try {
	    ret=(it->second)(opts);
	} catch (...) {
	    return 1;
	}
    } else {
	fprintf(stderr, PROGRAM ": invalid operation selected (%s)\n",cmd);
	return 1;
    }

    return ret;
}


int cmd_read_mem(Getopt& opts)
{
    bool have_prg=true;
    bool basic_load=false;

    if (opts.args_left()<1) {
	printf("too few arguments\n");
	return 1;
    }
    if (opts.args_left()<2) {
	basic_load=true;
    }

    const char* name;
    MemBuffer mb;
    Port pt(portbase_g);
    Transfer tf(pt,blocksize_g);

    if (!basic_load) {
	const char* range=opts.next_arg();
	name=opts.next_arg();

	mb.set_range(range);
    } else {
	name=opts.next_arg();
	uint8_t sabuf[2];
	uint8_t eabuf[2];
	uint32_t sa,ea;

	tf.read_mem(0,  0x002b, 2, sabuf );
	tf.read_mem(0,  0x002d, 2, eabuf );
	sa=sabuf[0] | (sabuf[1]<<8);
	ea=eabuf[0] | (eabuf[1]<<8);
	mb.set_start(sa);
	mb.set_length(ea-sa);
    }
	
    Timer tim;

    uint32_t len=mb.get_length();
    int blks=(len+2+253)/254;

    printf("transfering... $%06X-$%06X (%d bytes, %d %s)\n",mb.get_start(), mb.get_end(),mb.get_length(),blks,(blks==1)?"block":"blocks");
    fflush(stdout);
    tim.start();
    tf.read_mem( (mb.is_rom())?1:0,  mb.get_start(), mb.get_length(), mb.get_buffer() );
    tim.stop();

    float dur=static_cast<float>(tim.get_time_d());
    float speed_bytes=(len/dur);
    float speed_bits=(len/dur)*8.0f;
    printf ("transfered %d bytes in %f s (%f bytes/s, %f bit/s)\n",len,dur,speed_bytes,speed_bits);


    printf("writing file...\n");
    fflush(stdout);
    mb.write_file(name,!rawmode_g);

    return 0;
}

int cmd_write_mem(Getopt& opts)
{
    bool have_prg=!rawmode_g;
    bool set_endaddr=true;

    if (opts.args_left()<1) {
	printf("too few arguments\n");
	return 1;
    }
    
    const char* name;

    MemBuffer mb;
    if (opts.args_left()<2) {
	name=opts.next_arg();  
	mb.read_file(name,true);
    } else {
	const char* addrstr=opts.next_arg();
	name=opts.next_arg();  
	mb.read_file(name,have_prg);
	mb.set_start(addrstr);
    }

    uint32_t len=mb.get_length();
    int blks=(len+2+253)/254;

    Port pt(portbase_g);
    Transfer tf(pt,blocksize_g);
    Timer tim;

    printf("transfering... $%06X-$%06X (%d bytes, %d %s)\n",mb.get_start(), mb.get_end(),mb.get_length(),blks,(blks==1)?"block":"blocks");
    fflush(stdout);
    tim.start();
    tf.write_mem( (mb.is_rom())?1:0, mb.get_start(), mb.get_length(), mb.get_buffer() );

    if (set_endaddr) {
	uint32_t ea=mb.get_end();
	uint8_t abuf[2];
	abuf[0]=ea & 0xff;
	abuf[1]=(ea >> 8) & 0xff;
	tf.write_mem(0, 0x00ae, 2, abuf );
	tf.write_mem(0, 0x002d, 2, abuf );
    }

    tim.stop();
    float dur=static_cast<float>(tim.get_time_d());
    float speed_bytes=(len/dur);
    float speed_bits=(len/dur)*8.0f;
    printf ("transfered %d bytes in %f s (%f bytes/s, %f bit/s)\n",len,dur,speed_bytes,speed_bits);

    return 0;
}

int cmd_execute_mem(Getopt& opts)
{
    if (opts.args_left()<1) {
	printf("too few arguments\n");
	return 1;
    }
    const char* start=opts.next_arg();
    char* endptr;

    uint32_t startaddr=std::strtoul(start,&endptr,0);

    Port pt(portbase_g);
    Transfer tf(pt,blocksize_g);

    tf.execute_mem(startaddr);

    return 0;
}

int cmd_boot(Getopt& opts)
{
    if (opts.args_left()<1) {
	printf("too few arguments\n");
	return 1;
    }
    const char* name=opts.next_arg();

    MemBuffer mb;
    mb.read_file(name);

    uint32_t sa=mb.get_start();
    uint32_t len=mb.get_length();

    Port pt(portbase_g);
    Boot bt(pt);
    Timer tim;

    printf("sending boot... ($%04X-$%04X)\n",sa,sa+len-1);
    fflush(stdout);
    tim.start();
    bt.send_file(sa,sa+len-1, mb.get_buffer());
    tim.stop();
   
    float dur=tim.get_time_d();
    float speed_bytes=(len/dur);
    float speed_bits=(len/dur)*8.0f;
    printf ("sent %d bytes in %f s (%f bytes/s, %f bit/s)\n",len,dur,speed_bytes,speed_bits);

    return 0;
}

int cmd_clear_port(Getopt& opts)
{
    Port pt(portbase_g);
    pt.tristate();
    printf("Tri-stating port\n");
    return 0;
}

/* eof */
